from .basic_utils import *
import re
import time
import string
import random
Import("tqdm",globals())
Import("requests",globals())

def RandString(length:int, charset:str=string.ascii_uppercase + string.digits):
    """Return a random string.

    Args:
        length (int): Length of the random string.
        charset (str): The alphabet. Notice that if a word appears multiple times, the probability that it is chosen also multiples.
    Returns:
        str: A random string of length `length`, with all characters chosen from `charset`.
    """
    return ''.join(random.choice(charset) for _ in range(length))

def LineToFloats(line:str):
    """Read all float numbers from a line of text.

    Args:
        line (str): A line of text.
    Returns:
        List: All float numbers find in the line.
    """
    return [float(s) for s in re.findall(r"(?<!\w)[-+]?\d*\.?\d+(?!\d)",line)]

def FlattenList(L:List):
    """Flatten an embedded list to a 1D list.

    Args:
        L: The embedded list to be flattened.
    Returns:
        List: The flattened list.
    """
    F = lambda x:[e for i in x for e in F(i)] if isinstance(x,list) else [x]; return F(L)

def IP():
    """Return IP address string of the current device.

    Args:
        None
    Returns:
        str: The IP address string of the current device.
    """
    return requests.get('https://api.ipify.org').text

def FORMATTED_TIME(format="[%Y-%m-%d_%H.%M.%S]"):
    """Return the formatted current time.

    Args:
        format (str): The time format.
    Returns:
        str: The formatted current time.
    """
    return time.strftime(format,time.localtime(time.time()))

def Clipped(value, l, r):
    """Return a clipped value between lower bound `l` and upper bound `r`.

    Args:
        value: A value.
        l: The lower bound.
        u: The upper bound.
    Returns:
        Any: The clipped value.
    """
    return min(max(value,l),r)

def Unique(ls:List, **sort_args):
    """Return a sorted list of unique elements.

    Args:
        ls (List): The input list (or other iterable).
        sort_args: Args for calling `sorted` on the subpaths in format of (path, full_path), only works if `ordered` is True. Notice that the sort function will only be applied at the base level instead of applied recursively. PPlease refer to function `BUILTIN_LISTPATHS_SORT_CRITERIA()` for built-in criteria.

    Returns:
        List: The sorted list of unique elements.
    """
    return sorted(list(set(ls)), **sort_args)


from .file_utils import *
import io
import torch
import numpy as np
from torch.utils.data import Dataset, Subset, ConcatDataset, DataLoader

def SaveTorch(obj, path, pickle_protocol:Optional[int]=None, _use_new_zipfile_serialization:bool=True):
    """Save an object as torch file.

    Args:
        obj: The object to be saved.
        path: The save path.
        pickle_protocol (int/None): The `pickle_protocol` argument for `torch.save`.
        _use_new_zipfile_serialization (int/None): The `_use_new_zipfile_serialization` argument for `torch.save`.
    Returns:
        None
    """
    CreateFile(path); path = p2s(path)
    torch.save(obj, path, pickle_protocol=pickle_protocol, _use_new_zipfile_serialization=_use_new_zipfile_serialization)

def LoadTorch(path, map_location=None):
    """Load an object from existing torch file.

    Args:
        path: The load path.
        map_location: The `map_location` argument for `torch.load`.
    Returns:
        Any: The loaded object.
    """
    assert (ExistFile(path)), (f"Path '{path}' does not exist!"); path = p2s(path)
    return torch.load(path, map_location=map_location)

def DumpsTorch(obj, pickle_protocol:Optional[int]=None, _use_new_zipfile_serialization:bool=True):
    """Save an object as torch bytearray.

    Args:
        obj: The object to be saved.
        pickle_protocol (int/None): The `pickle_protocol` argument for `torch.save`.
        _use_new_zipfile_serialization (int/None): The `_use_new_zipfile_serialization` argument for `torch.save`.
    Returns:
        None
    """
    buff = io.BytesIO()
    torch.save(obj, buff, pickle_protocol=pickle_protocol, _use_new_zipfile_serialization=_use_new_zipfile_serialization)
    buff.seek(0); return buff.read()

class HeavenDataset(Dataset):
    def __init__(self, dataset, name="HeavenDataset"):
        self.dataset = dataset; self.name=name; self.__index__ = 0
    
    def __getitem__(self, i):
        if isinstance(i, int):
            return self.dataset[i]
        else:
            indices = np.array([j for j in range(len(self.dataset))])[i]
            return HeavenDataset(Subset(self, indices.tolist()),name=self.name)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]
    
    def __next__(self):
        if self.__index__>=len(self):
            raise StopIteration
        else:
            data = self[self.__index__]; self.__index__ += 1; return data

    def rename(self, name="HeavenDataset"):
        self.name = name
    
    def __len__(self):
        return len(self.dataset)
    
    def __or__(self, b):
        return HeavenDataset(ConcatDataset([self,b]),name=self.name+(","+b.name if hasattr(b,"name") else ""))

    def __str__(self):
        return ("{"+self.name+"}" if ',' in self.name else self.name)+f"(len={len(self)})"

class HeavenDataLoader:
    def __init__(self, dataloader, dataset=None):
        self.dataset = dataset
        self.dataloader = dataloader
        self.iterator = iter(dataloader)

    def __iter__(self):
        while True:
            for data in self.dataloader:
                yield data

    def __next__(self):
        try:
            data = next(self.iterator)
        except StopIteration:
            self.iterator = iter(self.dataloader)
            data = next(self.iterator)
        return data

    def __len__(self):
        return len(self.dataloader)


from .file_utils import *
from .misc_utils import Clipped
import PySide6

from PySide6 import QtWidgets
from PySide6.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QDialog,
    QFileDialog,
    QFrame,
    QGridLayout,
    QHBoxLayout,
    QInputDialog,
    QLabel,
    QLayout,
    QLineEdit,
    QListWidget,
    QMainWindow,
    QMessageBox,
    QProgressBar,
    QProgressDialog,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QSpinBox,
    QSpinBox,
    QStackedWidget,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from PySide6 import QtGui
from PySide6.QtGui import (
    QFont,
    QIcon,
)

from PySide6 import QtCore
from PySide6.QtCore import (
    QCoreApplication,
    QTimer,
    QRect,
    Qt,
)

from functools import partial

def QWrapped(widget):
    grid = QGridLayout()
    widget.setLayout(grid)
    widget.setMaximumSize(65536, 65536)
    widget.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
    grid.setContentsMargins(10, 10, 10, 10)
    grid.setSpacing(2)
    return widget

def QAdd(widget, module, pos, span, **args):
    instance = module(widget, **args)
    widget.layout().addWidget(instance, pos[0], pos[1], span[0], span[1])
    instance.setMaximumSize(65536, 65536)
    instance.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
    return instance

def QGetInput(base, title, question, default=None):
    answer, done = QInputDialog.getText(base, title, question)
    return answer if done else default

def QGetFolder(base, title, default):
    folder = QFileDialog.getExistingDirectory(base, title, default)
    return folder if folder!="" and ExistFolder(folder) else None

def QGetFile(base, title, default, filter="All Files (*)"):
    file, file_type = QFileDialog.getOpenFileName(base, title, default, filter=filter)
    return file if file!="" and file_type!="" and ExistFile(file) else None

def QGetNewFile(base, title, default="C:/", filter="All Files (*)"):
    file, file_type = QFileDialog.getSaveFileName(base, title, default, filter=filter)
    return file if file!="" and file_type!="" else None

class QWaiter(QProgressDialog):
    def __init__(self, base, title, min_dur=1000):
        super(QWaiter, self).__init__(base)
        self.setWindowTitle(title)
        self.setWindowModality(Qt.WindowModal)
        self.setCancelButton(None)
        self.setMinimumDuration(min_dur)
        self.setRange(0, 100)
        self.progress = 0
        self.base = base
    
    def refresh(self):
        self.setValue(self.progress)
        self.base.refresh()

def QWait(base, title, generator, min_dur=1000, default_label="", refresh=500, **args):
    waiter = QWaiter(base, title, min_dur, **args)
    waiter.setLabelText(default_label)
    waiter.show()
    timer = QTimer()
    timer.timeout.connect(waiter.refresh)
    timer.start(refresh)
    return_value = None
    for (progress,description,done,info) in generator:
        waiter.setLabelText(description)
        waiter.progress = Clipped(int(progress*100),0,99)
        waiter.refresh()
        if done:
            return_value = info; break
    timer.stop()
    waiter.progress = 100
    waiter.refresh()
    waiter.hide()
    waiter.close()
    return return_value

def QHint(base, title, message):
    hint = QMessageBox(base)
    hint.setWindowTitle(title)
    hint.setWindowModality(Qt.WindowModal)
    hint.setText(message)
    hint.exec()

def QConfirm(base, title, message):
    return (QMessageBox.question(base, title, message, QMessageBox.Yes | QMessageBox.No, QMessageBox.No) == QMessageBox.Yes)




from .misc_utils import RandString
from .file_utils import *
import hashlib
import string
import random
import base64

def Encrypt(bytes:bytearray, password:str=""):
	"""Encrypting a bytearray object with a password.

	This is a simple method of encryption, which does not guarantee security. Please do not encrypt important personal data via this.

	Args:
		bytes (bytearray): The bytearray to be encrypted.
		password (str): The password string.
	Returns:
		str: Return the cipher.
	"""
	random_state = random.getstate()
	h = hashlib.sha512(); h.update(password.encode("utf-8")); p = h.hexdigest(); assert(len(p)==128)
	b = list(base64.b64encode(bytes).decode('ascii'))
	base64charset = string.ascii_uppercase+string.ascii_lowercase+string.digits+"+/"
	for i in range(8):
		hexseed = p[i*16:(i+1)*16]; seed = int(hexseed, base=16); random.seed(seed)
		s = RandString(64,charset=base64charset); b.extend(s)
	for i in range(8):
		hexseed = p[i*16:(i+1)*16]; seed = int(hexseed, base=16)
		random.seed(seed); random.shuffle(b)
	random.setstate(random_state)
	return "".join(b)

def Decrypt(cipher:str, password:str=""):
	"""Decrypting a cipher, which was a bytearray object encrypted by `Encrypt` with the same password.

	Args:
		bytes (bytearray): The cipher to be decrypted.
		password (str): The password string.
	Returns:
		str: Return the decrypted bytearray object.
	"""
	try:
		random_state = random.getstate()
		h = hashlib.sha512(); h.update(password.encode("utf-8")); p = h.hexdigest(); assert(len(p)==128)
		k = [i for i in range(len(cipher))]; b = ["0" for _ in range(len(cipher))]
		for i in range(8):
			hexseed = p[i*16:(i+1)*16]; seed = int(hexseed, base=16)
			random.seed(seed); random.shuffle(k)
		for i,c in zip(k,cipher):
			b[i] = c
		random.setstate(random_state)
		b = ("".join(b))[:-512].encode('ascii')
		return base64.b64decode(b)
	except:
		raise Exception("Incorrect password or corrupted cipher!")
